package explorviz.hpc_monitoring.reader;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import com.lmax.disruptor.EventHandler;
import com.lmax.disruptor.RingBuffer;
import com.lmax.disruptor.dsl.Disruptor;

import explorviz.hpc_monitoring.byteaccess.UnsafeBits;
import explorviz.hpc_monitoring.filter.counting.CountingThroughputFilter;
import explorviz.hpc_monitoring.filter.reconstruction.TraceReconstructionFilter;
import explorviz.hpc_monitoring.record.IRecord;
import explorviz.hpc_monitoring.record.TraceMetadata;
import explorviz.hpc_monitoring.record.events.normal.AfterFailedOperationEvent;
import explorviz.hpc_monitoring.record.events.normal.AfterOperationEvent;
import explorviz.hpc_monitoring.record.events.normal.BeforeOperationEvent;
import gnu.trove.map.hash.TIntObjectHashMap;

public class MessageDistributer implements EventHandler<ByteArrayEvent> {

	private static final CountingThroughputFilter counter = new CountingThroughputFilter(
			"Records per second");

	private final TIntObjectHashMap<String> stringRegistry = new TIntObjectHashMap<String>(
			64);

	private byte[] unreadBytes = null;

	private final RingBuffer<RecordEvent> ringBuffer;

	@SuppressWarnings("unchecked")
	public MessageDistributer(final EventHandler<RecordEvent> endReceiver) {
		final ExecutorService exec = Executors.newCachedThreadPool();
		final Disruptor<RecordEvent> disruptor = new Disruptor<RecordEvent>(
				RecordEvent.EVENT_FACTORY, 32768, exec);

		final EventHandler<RecordEvent>[] eventHandlers = new EventHandler[1];
		eventHandlers[0] = new TraceReconstructionFilter(
				1 * 1000 * 1000 * 1000, endReceiver);
		disruptor.handleEventsWith(eventHandlers);
		ringBuffer = disruptor.start();
	}

	@Override
	public void onEvent(final ByteArrayEvent event, final long sequence,
			final boolean endOfBatch) throws Exception {
		final byte[] received = event.getValue();
		final int receivedLength = event.getLength();

		byte[] messages = received;
		int messagesLength = receivedLength;

		if (unreadBytes != null) {
			final int unreadBytesLength = unreadBytes.length;

			messagesLength += unreadBytesLength;
			messages = new byte[messagesLength];

			System.arraycopy(unreadBytes, 0, messages, 0, unreadBytesLength);
			System.arraycopy(received, 0, messages, unreadBytesLength,
					receivedLength);
		}

		unreadBytes = messagesfromByteArray(messages, messagesLength);
	}

	private byte[] messagesfromByteArray(final byte[] b, final int readSize) {
		int offset = 0;

		while (offset < readSize) {
			if ((readSize - offset) < 4) {
				return createUnreadBytesArray(b, readSize, offset, false);
			}

			final int clazzId = UnsafeBits.getInt(b, offset);
			offset += 4;

			IRecord record = null;

			switch (clazzId) {
			case 0: {
				if ((readSize - offset) < (8 + 4 + 8 + 4 + 4)) {
					return createUnreadBytesArray(b, readSize, offset, true);
				}

				final long traceId = UnsafeBits.getLong(b, offset);
				offset += 8;
				final int hostnameId = UnsafeBits.getInt(b, offset);
				offset += 4;
				final long parentTraceId = UnsafeBits.getLong(b, offset);
				offset += 8;
				final int parentOrderId = UnsafeBits.getInt(b, offset);
				offset += 4;
				final int applicationId = UnsafeBits.getInt(b, offset);
				offset += 4;

				record = new TraceMetadata(traceId,
						getStringFromRegistry(hostnameId), parentTraceId,
						parentOrderId, getStringFromRegistry(applicationId));
				break;
			}
			case 1: {
				if ((readSize - offset) < (8 + 8 + 4 + 4)) {
					return createUnreadBytesArray(b, readSize, offset, true);
				}

				final long timestamp = UnsafeBits.getLong(b, offset);
				offset += 8;
				final long traceId = UnsafeBits.getLong(b, offset);
				offset += 8;
				final int orderIndex = UnsafeBits.getInt(b, offset);
				offset += 4;
				final int operationId = UnsafeBits.getInt(b, offset);
				offset += 4;

				record = new BeforeOperationEvent(timestamp, traceId,
						orderIndex, getStringFromRegistry(operationId));
				break;
			}
			case 2: {
				if ((readSize - offset) < (8 + 8 + 4 + 4 + 4)) {
					return createUnreadBytesArray(b, readSize, offset, true);
				}

				final long timestamp = UnsafeBits.getLong(b, offset);
				offset += 8;
				final long traceId = UnsafeBits.getLong(b, offset);
				offset += 8;
				final int orderIndex = UnsafeBits.getInt(b, offset);
				offset += 4;
				final int operationId = UnsafeBits.getInt(b, offset);
				offset += 4;
				final int causeId = UnsafeBits.getInt(b, offset);
				offset += 4;

				record = new AfterFailedOperationEvent(timestamp, traceId,
						orderIndex, getStringFromRegistry(operationId),
						getStringFromRegistry(causeId));
				break;
			}
			case 3: {
				if ((readSize - offset) < (8 + 8 + 4 + 4)) {
					return createUnreadBytesArray(b, readSize, offset, true);
				}

				final long timestamp = UnsafeBits.getLong(b, offset);
				offset += 8;
				final long traceId = UnsafeBits.getLong(b, offset);
				offset += 8;
				final int orderIndex = UnsafeBits.getInt(b, offset);
				offset += 4;
				final int operationId = UnsafeBits.getInt(b, offset);
				offset += 4;

				record = new AfterOperationEvent(timestamp, traceId,
						orderIndex, getStringFromRegistry(operationId));
				break;
			}
			case 4: {
				if ((readSize - offset) < (4 + 4)) {
					return createUnreadBytesArray(b, readSize, offset, true);
				}

				final int mapId = UnsafeBits.getInt(b, offset);
				offset += 4;
				final int stringLength = UnsafeBits.getInt(b, offset);
				offset += 4;

				if ((readSize - offset) < stringLength) {
					return createUnreadBytesArray(b, readSize, offset - 8, true);
				}

				final byte[] stringBytes = new byte[stringLength];
				System.arraycopy(b, offset, stringBytes, 0, stringLength);
				final String string = new String(stringBytes);
				offset += stringLength;

				addToRegistry(mapId, string);

				break;
			}
			default: {
				System.out.println("unknown class id " + clazzId
						+ " at offset " + (offset - 4));
				return null;
			}
			}

			putInRingBuffer(record);
		}

		return null;
	}

	private byte[] createUnreadBytesArray(final byte[] b, final int readSize,
			int offset, final boolean withClazzId) {
		if (withClazzId) {
			offset -= 4;
		}
		final int unreadBytesSize = readSize - offset;
		final byte[] unreadBytes = new byte[unreadBytesSize];
		System.arraycopy(b, offset, unreadBytes, 0, unreadBytesSize);
		return unreadBytes;
	}

	private void putInRingBuffer(final IRecord record) {
		counter.inputObjects(record);
		final long hiseq = ringBuffer.next();
		final RecordEvent valueEvent = ringBuffer.get(hiseq);
		valueEvent.setValue(record);
		ringBuffer.publish(hiseq);
	}

	public void addToRegistry(final int key, final String value) {
		stringRegistry.put(key, value);
		
		System.out.println("put key " + key + " value " + value);

		synchronized (this) {
			notifyAll();
		}
	}

	private String getStringFromRegistry(final int id) {
		String result = stringRegistry.get(id);
		while (result == null) {
			try {
				synchronized (this) {
					System.out.println("waiting for " + id);
					this.wait();
				}
			} catch (final InterruptedException e) {
				e.printStackTrace();
			}
			result = stringRegistry.get(id);
		}

		return result;
	}
}
