package explorviz.hpc_monitoring.reader;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import com.lmax.disruptor.EventHandler;
import com.lmax.disruptor.RingBuffer;
import com.lmax.disruptor.dsl.Disruptor;

public final class TCPReader {
	private static final int MESSAGE_BUFFER_SIZE = 65536;

	private final int listeningPort;

	private ServerSocket serversocket;
	private boolean active = true;

	private final RingBuffer<ByteArrayEvent> ringBuffer;

	public TCPReader(final int listeningPort,
			final EventHandler<RecordEvent> endReceiver)
			throws IllegalArgumentException {
		this.listeningPort = listeningPort;

		final ExecutorService exec = Executors.newCachedThreadPool();
		final Disruptor<ByteArrayEvent> disruptor = new Disruptor<ByteArrayEvent>(
				ByteArrayEvent.EVENT_FACTORY, 4096, exec);

		@SuppressWarnings("unchecked")
		final EventHandler<ByteArrayEvent>[] eventHandlers = new EventHandler[1];
		eventHandlers[0] = new MessageDistributer(endReceiver);
		disruptor.handleEventsWith(eventHandlers);
		ringBuffer = disruptor.start();
	}

	public void read() {
		try {
			open();
			while (active) {
				// TODO only one connection!
				final Socket socket = serversocket.accept();
				final BufferedInputStream bufferedInputStream = new BufferedInputStream(
						socket.getInputStream(), MESSAGE_BUFFER_SIZE);
				int readBytes = 0;
				byte[] messages = new byte[MESSAGE_BUFFER_SIZE];
				while ((readBytes = bufferedInputStream.read(messages, 0,
						MESSAGE_BUFFER_SIZE)) != -1) {
					putInRingBuffer(messages, readBytes);
					messages = new byte[MESSAGE_BUFFER_SIZE];
				}

				socket.close();
			}
		} catch (final IOException ex) {
			System.out.println("Error in read() " + ex.toString());
		} finally {
			try {
				serversocket.close();
			} catch (final IOException e) {
				System.out.println("Error in read()" + e.toString());
			}
		}
	}

	private void open() throws IOException {
		serversocket = new ServerSocket(listeningPort);
		System.out.println("listening on port " + listeningPort);
	}

	private void putInRingBuffer(final byte[] messages, final int readBytes) {
		final long hiseq = ringBuffer.next();
		final ByteArrayEvent valueEvent = ringBuffer.get(hiseq);
		valueEvent.setValue(messages);
		valueEvent.setLength(readBytes);
		ringBuffer.publish(hiseq);
	}

	public void terminate(final boolean error) {
		System.out.println("Shutdown of TCPReader requested.");
		active = false;
	}
}
